/*
 * u8_functions.cpp
 * 
 * Musterloesung zu Uebung 8, Informatik I, D-ITET, WS 03/04
 *
 * Edouard Lamboray, 17.12.2002
 */


#include "u8_functions.h" 
#include <stdlib.h>
#include <fstream>
#include <iostream>
#include <iomanip>
using namespace std;

/////////////////////////////////////////////////////////////
//              The little/big endian problem              //
/////////////////////////////////////////////////////////////
//
// from www.webopedia.com
/*
  The adjectives big-endian and little-endian
  refer to which bytes are most significant in
  multi-byte data types and describe the order
  in which a sequence of bytes is stored in a
  computer's memory. 
  
  In a big-endian system, the most significant
  value in the sequence is stored at the lowest
  storage address (i.e., first). In a little-endian
  system, the least significant value in the
  sequence is stored first. For example,
  consider the number 1025 (2 to the tenth
  power plus one) stored in a 4-byte integer: 
  
  00000000 00000000 00000100 00000001 

  Address  Big-Endian      Little-Endian
           representation  representation
           of 1025         of 1025 

  00       00000000        00000001
  01       00000000        00000100
  02       00000100        00000000
  03       00000001        00000000 


  The terms big-endian and little-endian are derived 
  from the Lilliputians of Gulliver's Travels, whose
  major political issue was whether soft-boiled eggs 
  should be opened on the big side or the little side.
  Likewise, the big-/little-endian computer debate has 
  much more to do with political issues than
  technological merits.
*/
/*
  A binary file becomes readable on any machine if its
  endianess is defined and if the appropriate transforms
  are applied while reading and writing the data.
  Note that similar problems occur if data is transmitted
  via a communication network, i.e. the Internet. In 
  general, the terms network and host byte order are used.
  
  In this exercise, we use the big endian format for 
  storing the data.
*/

// Swaps the byte order of a four byte word on little endian machines. 
// On big endian machines, the bytes are simply copied from in to out.
void swapEndian4(char *in, char *out) 
{
  // Test the endianess
  long l=1; 
  if(*(char *)&l == 0) {
    // Big endian, copy the word and keep the byte order
    out[0]=*(in);
    out[1]=*(in+1);
    out[2]=*(in+2);
    out[3]=*(in+3);
  }
  else {
    // Little endian, copy the word and change the byte order
    out[0]=*(in+3);
    out[1]=*(in+2);
    out[2]=*(in+1);
    out[3]=*(in);
  }
}

// Helper function for coding a date into the prescribed format.
int codeDate(int day, int month, int year) {
  int i = day << 24;
  i |= (month << 16);
  i |= year;
  return i;
}

// Helper function for decoding a date from the prescribed format.
void decodeDate(int d, unsigned short &day, unsigned short &month, 
                unsigned short &year) {
  day = (d & 0xFF000000) >> 24;
  month = (d & 0x00FF0000) >> 16;
  year = d & 0x0000FFFF;
}

// Returns the number of data items of the predefined format for (date, 
// temperature) values in a file.
// Returns -1 on error.
int numberOfItems(const char *filename)
{
  int counter = 0;

  ifstream f(filename);  
  if(!f.is_open()) return -1; // file not found

  while(!f.eof()) {
    f.get();
    counter++;
  }
  f.close();
  return counter/(sizeof(int) + sizeof(float));
}

// Reads the data items of the predefined format for (date, temperature) from
// a file into the array of Temperature structs.
// The Temperature array must be allocated and must be of sufficient size.
// Returns the number of correctly read data items.
// Returns -1 on error.
int readTemperatures(const char *filename, Temperature *t, int size) 
{
  ifstream f(filename);  
  if(!f.is_open()) return -1; // file not found

  int dn, dh;
  float xn, xh;

  for(int i=0; i<size; i++) {

    // Read the data
    f.read((char *) &dn, sizeof dn);
    f.read((char *) &xn, sizeof xn);

    // Check if the read was correct.
    if(f.bad() | f.fail() | f.eof()) return i;

    // Transform d from the network byte order to the host byte order
    // Solves the little/big endian problem.
    swapEndian4((char *)&dn, (char *)&dh);
    swapEndian4((char *)&xn, (char *)&xh);

    // Set the Temperature struct.    
    decodeDate(dh, t[i].day, t[i].month, t[i].year);
    t[i].temperature = xh;
  }

  return size; 
}

// Writes the data items from an array of Temperature structs into a file 
// using the predefined format for (date, temperature).
// if t==0, the temperatures are randomly generated.
// Returns the number of correctly written data items.
// Returns -1 on error.
int writeTemperatures(const char *filename, Temperature *t, int size) 
{
  // Opening the file
  ofstream f(filename);
  if(t == 0) {
    cout << "Generating " << size << " random temperatures for January 2004.\n";
    if(size < 1 || size > 31) size = 31;
    for(int i=1; i<=size; i++) {
      const float th = ((unsigned char)(rand())/30.0) -4.0;
      const int xh = codeDate(i,1,2004);

      // Transform x from the host byte order to the network byte order
      // Solves the little/big endian problem.
      int xn;
      float tn;
      swapEndian4((char *)&xh, (char *)&xn);
      swapEndian4((char *)&th, (char *)&tn);

      f.write((char *) &xn, sizeof xn);
      f.write((char *) &tn, sizeof tn);
    }
  }
  else {
    // normal procedure
    for(int i=1; i<=size; i++) {
      const int xh = codeDate(t[i].day, t[i].month, t[i].year);

      // Transform x from the host byte order to the network byte order
      // Solves the little/big endian problem.
      int xn; 
      float tn;
      swapEndian4((char *)&xh, (char *)&xn);      
      swapEndian4((char *)&t[i].temperature, (char *)&tn);

      f.write((char *) &xn, sizeof xn);
      f.write((char *) &tn, sizeof tn);
    }
  }

  // Closing the file
  f.close();
  return size;
}

// Writes the data items from an array of Temperature structs into a text file.
// The last line in the text file contains the average temperature of the data
// set.
// Returns the number of correctly written data items.
// Returns -1 on error.
int writeTemperaturesAsText(const char *filename, Temperature *t, 
                            int size, float average) 
{
  ofstream f(filename);
  if(!f.is_open()) return -1;

  f.precision(1);                         // floating point display precision
  f.setf(ios::fixed, ios::floatfield);    // fixed format

  for(int i=0; i<size; i++) {
    f.width(2);  
    f << t[i].day << "." << t[i].month << "." << t[i].year << ":\t";
    f.width(5);
    f << t[i].temperature << endl;
  }
  f << setw(24) << setfill('-') << "" << setfill(' ') << endl;
  f << "Schnitt:\t" << setw(5) << average << endl;
  f.close();
  return size;    
}

// Computes the average temperature of a data set.
float computeAverage(Temperature *t, int size)
{
  float a = 0.0;
  for(int i=0; i<size; i++) {
    a += t[i].temperature;
  }
  return a/size;
}

